# -*- coding: utf-8 -*-
"""
Created on Tue Sep 21 09:44:36 2021

@author: RNDr. Andrej Krafcik, PhD. 
         Department of Imaging Methods, Institute of Measurement Science, 
         Slovak Academy of Sciences, Bratislava, Slovakia.
@environment: Python 3.8.3

Magnetic particle alignment models comparison:
    notation    torques involved
    SDE         inertial, magnetic, viscous quasi-steady, stochastic Brownian
    SIDE        inertial, magnetic, viscous quasi-steady, viscous history
                acceleration, stochastic Brownian
    ODE         inertial, magnetic, viscous quasi-steady
    IDE         inertial, magnetic, viscous quasi-steady, viscous history
                accelaration
In further combination of viscous quasi-steady and viscous history acceleration
torques are denoted as arbitrary viscous torque.
    
Description: 
    Numerical simulation of magnetic particle alignment process in external 
    magnetic field with flux density B_0 and water as ambient fluid. 
    Comparison of four different models, each involving different combination 
    of considered torques. As the particle were considered properties of 
    commertially available MyOne 1um (ThermoFischer Scientific, DynabeadsTM, 
    MyOneTM) in constant magnetic moment magnitude simplification.
    
Software Source Code as a Supplementary Material to the manuscript:
    Krafcik A, Babinec P, Strbak O, Frollo I (2021): 'A theoretical analysis
    of magnetic particle alignment in external magnetic fields affected by 
    viscosity and Brownian motion." Submitted to the Applied Sciences journal.
"""

import numpy as np
import matplotlib.pyplot as plt
from math import pi, exp, sqrt, erfc, sin
from scipy.io import savemat


## Define problems:
h = 1e-8                # in s; Finite time-step
B_0 = 5e-0              # in T; External magnetic flux density
t_0 = 0.0               # in s; Initial time
t_end = 1e-6            # in s; End time

phi_0 = pi/2            # in rad; Initial magnetic particle alignment angle
omega_0 = 0             # in rad/s; Initial magnetic particle alignment angular velocity

T = 293.15              # in K; Thermodynamic temperature of ambient fluid
k_B = 1.38065e-23       # in J/K; Boltzmann constant
eta = 1e-3              # in Pa.s; Fluid dynamic viscosity
rho = 1e3               # in kg/m3; Fluid mass density
nu = eta/rho            # in m2/s; Fluid kinematic viscosity
R = 0.5e-6              # in m; Particle radius
V_p = 4/3*pi*R**3       # in m3; Particle volume
rho_p = 1792            # in kg/m3; Particle mass density
m_p = V_p*rho_p         # in kg; Particle weight
I_p = 2/5*m_p*R**2      # in kg.m2; Particle moment of inertia
M_sp = 43e3             # in A/m; Particle saturation magnetization
mu_p = M_sp*V_p         # in A.m2; Particle magnetic moment
gamma = 8*pi*eta*R**3   # in kg.m2/s; Stokes coefficient

settings = {"h": h, "B_0": B_0, "t_0": t_0, "t_end": t_end, 
            "phi_0": phi_0, "omega_0": omega_0, 
            "T": T, "k_B": k_B,
            "eta": eta, "rho": rho, "nu": nu, 
            "R": R, "V_p": V_p, "rho_p": rho_p, "m_p": m_p,
            "I_p": I_p, "M_sp": M_sp, "mu_p": mu_p, "gamma": gamma}

print("Magnetic particle alignment problems settings:")
for key in settings:
    print(f'\t{key:8s} = {settings[key]:.5g}')


def solve_ode():
    # Solve ordinary differential equations (ODE) model involving: inertial,
    # magnetic, quasi-steady viscous, and no stochastic Brownian torques.
    
    tau = [t_0]
    phi = [phi_0]
    omega = [omega_0]                
    
    t = t_0
    while t<t_end:
        
        t += h
        tau.append(t)

        phi_im1 = phi[-1]
        if len(phi)<2:
            phi_im2 = phi_0
        else:
            phi_im2 = phi[-2]
        phi_i = (-B_0*h**2*mu_p*sin(phi_im1) + 2*I_p*phi_im1 - I_p*phi_im2 + gamma*h*phi_im1)/(I_p + gamma*h)
        omega_i = -(phi_i - phi_im1)/h
        
        phi.append(phi_i)
        omega.append(omega_i)
        
    sol ={"x": tau, "y": np.concatenate([[phi], [omega]]),
          "notation": "ODE",
          "describe": "inertial, magnetic, quasi-steady viscous, and no stochastic Brownian torques"}
    return sol


def solve_sde():
    # Solve stochastic differential equations (SDE) model involving: inertial,
    # magnetic, quasi-steady viscous, and stochastic Brownian torques.

    tau = [t_0]
    phi = [phi_0]
    omega = [omega_0]
    
    t = t_0
    while t<t_end:
        
        t += h
        tau.append(t)
        
        phi_im1 = phi[-1]
        if len(phi)<2:
            phi_im2 = phi_0
        else:
            phi_im2 = phi[-2]
        w_i = np.random.randn()
        phi_i = (-B_0*h**2*mu_p*sin(phi_im1) + 2*I_p*phi_im1 - I_p*phi_im2 - sqrt(2)*sqrt(T)*sqrt(gamma)*h**(3/2)*sqrt(k_B)*w_i + gamma*h*phi_im1)/(I_p + gamma*h)
        omega_i = -(phi_i - phi_im1)/h
        
        phi.append(phi_i)
        omega.append(omega_i)
        
    sol ={"x": tau, "y": np.concatenate([[phi], [omega]]),
          "notation": "SDE",
          "describe": "inertial, magnetic, quasi-steady viscous, and stochastic Brownian torques"}          
    return sol


def solve_ide():
    # Solve integro-differential equations (IDE) model involving: inertial,
    # magnetic, arbitrary viscous, and no stochastic Brownian torques.
    
    tau = [t_0]
    phi = [phi_0]
    omega = [omega_0]
    
    def I_i(t, i):
        if i==0:
            return 0.0
        else:
            return omega[i-1]*sqrt(pi)*R**2/nu * (exp(nu*(t-tau[i])/R**2) * erfc(sqrt(nu*(t-tau[i]))/R) \
                                                  - exp(nu*(t-tau[i-1])/R**2) * erfc(sqrt(nu*(t-tau[i-1]))/R)) \
                + (omega[i] - omega[i-1])/h* (2*R**3/nu**(3/2) * (sqrt(t - tau[i]) - sqrt(t - tau[i-1])) \
                                                                 + sqrt(pi)*(R**4/nu**2 + R**2/nu*h) * exp(nu*(t - tau[i])/R**2) * erfc(sqrt(nu*(t - tau[i]))/R) \
                                                                     - sqrt(pi)*(R**4/nu**2) * exp(nu*(t - tau[i-1])/R**2) * erfc(sqrt(nu*(t - tau[i-1]))/R) )
    
    t = t_0
    dIdt, Integral, integral = 0, 0, 0
    while t<t_end:
        
        Integral = 0
        for n in range(len(phi)):
            try:
                Integral += I_i(t, n)
            except:
                print(f'\tPremature stopping of integration at t = {t:g} s due to math range error in f(x) = exp(x).')
                sol ={"x": tau, "y": np.concatenate([[phi], [omega]]), 
                      "notation": "IDE",
                      "describe": "inertial, magnetic, arbitrary viscous, and no stochastic Brownian torques"}
                return sol
            
        dIdt = (Integral - integral) / h
        T_acc = - gamma/(3*sqrt(pi)) * dIdt
        integral = Integral
            
        phi_im1 = phi[-1]
        if len(phi)<2:
            phi_im2 = phi_0
        else:
            phi_im2 = phi[-2]
        phi_i = (-B_0*h**2*mu_p*sin(phi_im1) + 2*I_p*phi_im1 - I_p*phi_im2 - T_acc*h**2 + gamma*h*phi_im1)/(I_p + gamma*h)
        omega_i = -(phi_i - phi_im1)/h
        
        t += h
        tau.append(t)
        phi.append(phi_i)
        omega.append(omega_i)
        
    sol ={"x": tau, "y": np.concatenate([[phi], [omega]]), 
          "notation": "IDE",
          "describe": "inertial, magnetic, arbitrary viscous, and no stochastic Brownian torques"}
    return sol

def solve_side():
    # Solve stochastic integro-differential equations (SIDE) model involving: 
    # inertial,magnetic, arbitrary viscous, and stochastic Brownian torques.

    tau = [t_0]
    phi = [phi_0]
    omega = [omega_0]
    
    def I_i(t, i):
        if i==0:
            return 0.0
        else:
            return omega[i-1]*sqrt(pi)*R**2/nu * (exp(nu*(t-tau[i])/R**2) * erfc(sqrt(nu*(t-tau[i]))/R) \
                                                  - exp(nu*(t-tau[i-1])/R**2) * erfc(sqrt(nu*(t-tau[i-1]))/R)) \
                + (omega[i] - omega[i-1])/h* (2*R**3/nu**(3/2) * (sqrt(t - tau[i]) - sqrt(t - tau[i-1])) \
                                                                 + sqrt(pi)*(R**4/nu**2 + R**2/nu*h) * exp(nu*(t - tau[i])/R**2) * erfc(sqrt(nu*(t - tau[i]))/R) \
                                                                     - sqrt(pi)*(R**4/nu**2) * exp(nu*(t - tau[i-1])/R**2) * erfc(sqrt(nu*(t - tau[i-1]))/R) )
    
    t = t_0
    dIdt, Integral, integral = 0, 0, 0
    while t<t_end:
        
        Integral = 0
        for n in range(len(phi)):
            try:
                Integral += I_i(t, n)
            except:
                print(f'\tPremature stopping of integration at t = {t:g} s due to math range error in f(x) = exp(x).')
                sol ={"x": tau, "y": np.concatenate([[phi], [omega]]), 
                      "notation": "SIDE",
                      "describe": "inertial, magnetic, arbitrary viscous, and stochastic Brownian torques"}
                return sol  
            
        dIdt = (Integral - integral) / h
        T_acc = - gamma/(3*sqrt(pi)) * dIdt
        integral = Integral
            
        phi_im1 = phi[-1]
        if len(phi)<2:
            phi_im2 = phi_0
        else:
            phi_im2 = phi[-2]
        w_i = np.random.randn()
        phi_i = (-B_0*h**2*mu_p*sin(phi_im1) + 2*I_p*phi_im1 - I_p*phi_im2 - T_acc*h**2 - sqrt(2)*sqrt(T)*sqrt(gamma)*h**(3/2)*sqrt(k_B)*w_i + gamma*h*phi_im1)/(I_p + gamma*h)
        omega_i = -(phi_i - phi_im1)/h
        
        t += h
        tau.append(t)
        phi.append(phi_i)
        omega.append(omega_i)
        
    sol ={"x": tau, "y": np.concatenate([[phi], [omega]]), 
          "notation": "SIDE",
          "describe": "inertial, magnetic, arbitrary viscous, and stochastic Brownian torques"}
    return sol


## Solve problems:
print("Solving SDE problem...");   sde = solve_sde();  print("\tDONE.")  
print("Solving SIDE problem..."); side = solve_side(); print("\tDONE.")
print("Solving ODE problem...");   ode = solve_ode();  print("\tDONE.")
print("Solving IDE problem...");   ide = solve_ide();  print("\tDONE.")


## Show results:
print("Plotting results...")
LineWidth = 3

plt.figure(figsize=(10, 8))
plt.subplot(2, 1, 1)
handle = plt.plot(sde["x"], sde["y"][0, :], "-",
                  side["x"], side["y"][0, :], "-",
                  ode["x"], ode["y"][0, :], "--",
                  ide["x"], ide["y"][0, :], "--")
plt.setp(handle, "linewidth", LineWidth)
plt.title((f"$B_0 = ${B_0*1e3:g} mT, $T = ${T:g} K, $h = ${h:g} s"))
plt.ylabel("$\phi$ [rad]")
plt.legend(handle, ("SDE", "SIDE", "ODE", "IDE"))
plt.subplot(2, 1, 2)
handle = plt.plot(sde["x"], sde["y"][1, :], "-",
                  side["x"], side["y"][1, :], "-",
                  ode["x"], ode["y"][1, :], "-",
                  ide["x"], ide["y"][1, :], "--")
plt.setp(handle, "linewidth", LineWidth)
plt.xlabel("$t$ [s]")
plt.ylabel("$\omega$ [rad/s]")

plt.show()
print("\tDONE.")


## Save results as MATLAB .mat file:
matfile = "particle_alignment__{:08.3f}mT.mat".format(B_0*1e3)
print(f"Saving results as MATLAB '{matfile:s}' file...")
variables = {"settings": settings, 
             "sde": sde, "side": side, "ode": ode, "ide": ide}

savemat(matfile, variables)
print("\tDONE.")

print("DONE.")

